"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var _ = require("lodash");
var product_1 = require("../product/product");
var CartItem = (function () {
    function CartItem(product, quantity, courseId, size, colour, printer) {
        this.product = product;
        this.quantity = quantity;
        this.courseId = courseId;
        this.size = size;
        this.colour = colour;
        this.printer = printer;
        this.cost = this.product.getPrice() * this.quantity;
    }
    CartItem.fromJson = function (obj) {
        return new CartItem(product_1.Product.inCartFromJson(obj), obj.quantity, obj.courseId ? obj.courseId : null, obj.size ? obj.size : null, obj.colour ? obj.colour : null, product_1.Product.inCartFromJson(obj).getPrinter());
    };
    CartItem.prototype.toJson = function () {
        var cartItem = this.product.toJson();
        cartItem.quantity = this.getQuantity();
        if (this.courseId)
            cartItem.courseId = this.courseId;
        if (this.printer)
            cartItem.printer = this.printer;
        return cartItem;
    };
    CartItem.prototype.updateItemQuantity = function (updatedQuantity) {
        this.quantity = updatedQuantity;
        this.cost = this.product.getPrice() * this.quantity;
    };
    CartItem.prototype.getCost = function () { return this.cost; };
    CartItem.prototype.getProduct = function () { return this.product; };
    CartItem.prototype.getQuantity = function () { return this.quantity; };
    CartItem.prototype.getSize = function () { return this.size; };
    CartItem.prototype.getColour = function () { return this.colour; };
    CartItem.prototype.getCourseId = function () { return this.courseId ? this.courseId : null; };
    CartItem.prototype.getPrinter = function () { return this.printer ? this.printer : null; };
    return CartItem;
}());
exports.CartItem = CartItem;
var Cart = (function () {
    function Cart(numItems, storeId, tenantId, userId, cartItems, status, id, tableId, created, __v, _rev) {
        this.storeId = storeId;
        this.tenantId = tenantId;
        this.userId = userId;
        this.cartItems = cartItems;
        this.status = status;
        this.id = id;
        this.tableId = tableId;
        this.created = created;
        this.__v = __v;
        this._rev = _rev;
        this.cartCost = 0;
        this.numItems = (_.isUndefined(numItems) || numItems == null ? 0 : numItems);
        this.calculateCartCost();
        if (this.id === undefined) {
            this.id = '';
        }
    }
    Cart.fromJson = function (obj) {
        var items = _.chain(obj.items).values().map(CartItem.fromJson).value();
        return new Cart(Number(obj.numItems), obj.storeId, obj.tenantId, obj.userId, items, obj.status, obj.id, obj.tableId, obj.created, obj.__v, obj._rev);
    };
    Cart.prototype.toJson = function () {
        var cartJson = {
            __v: this.__v ? this.__v : '',
            userId: this.userId,
            tenantId: this.tenantId,
            id: this.id,
            created: this.created ? this.created : '',
            items: [],
            numItems: this.getItemsCount(),
            total: this.getCartCost(),
            storeId: this.storeId,
            status: this.status,
            tableId: this.tableId
        };
        if (this._rev)
            cartJson._rev = this._rev;
        return cartJson;
    };
    Cart.prototype.calculateCartCost = function () {
        var _this = this;
        this.cartCost = 0;
        this.cartItems.forEach(function (cItem) {
            _this.cartCost += cItem.getCost();
        });
    };
    Cart.prototype.findItemInCartItems = function (keyItem) {
        return _.find(this.cartItems, function (cItem) {
            return cItem.getProduct().getId() === keyItem.getProduct().getId();
        });
    };
    Cart.prototype.findProductInCartItems = function (keyId) {
        return _.find(this.cartItems, function (cItem) {
            return cItem.getProduct().getId() === keyId;
        });
    };
    Cart.prototype.addCartItem = function (newItem) {
        var foundItem = this.findItemInCartItems(newItem);
        var position = this.cartItems.findIndex(function (item) { return item.getProduct().getId() === newItem.getProduct().getId(); });
        this.numItems += newItem.getQuantity();
        if (foundItem) {
            this.cartItems[position].updateItemQuantity(foundItem.getQuantity() + newItem.getQuantity());
        }
        else {
            this.cartItems.push(newItem);
        }
        this.calculateCartCost();
    };
    Cart.prototype.removeCartItem = function (idOfProductToRemove) {
        var indexOfItemToRemove = _.findIndex(this.cartItems, function (cItem) {
            return cItem.getProduct().getId() === idOfProductToRemove;
        });
        if (indexOfItemToRemove === -1) {
            console.error('Failed to remove product with id "'
                + idOfProductToRemove + '". Product was not found in cartItems[].');
        }
        else {
            this.numItems -= this.cartItems[indexOfItemToRemove].getQuantity();
            this.cartItems.splice(indexOfItemToRemove, 1);
            this.calculateCartCost();
        }
    };
    Cart.prototype.getItem = function (prodId) {
        var indexOfItem = _.findIndex(this.cartItems, function (cItem) {
            return cItem.getProduct().getId() === prodId;
        });
        return this.cartItems[indexOfItem];
    };
    Cart.prototype.incrementQuantity = function (prodId) {
        var foundItem = this.findProductInCartItems(prodId);
        foundItem.updateItemQuantity(foundItem.getQuantity() + 1);
        this.setItemsCount();
        this.calculateCartCost();
        return foundItem;
    };
    Cart.prototype.decrementQuantity = function (prodId) {
        var foundItem = this.findProductInCartItems(prodId);
        foundItem.updateItemQuantity(foundItem.getQuantity() - 1);
        this.setItemsCount();
        this.calculateCartCost();
        return foundItem;
    };
    Cart.prototype.setItemsCount = function () { this.numItems = this.getCartItems().map(function (a) { return a.getQuantity(); }).reduce(function (a, b) { return a + b; }, 0); };
    Cart.prototype.getItemsCount = function () { return this.numItems; };
    Cart.prototype.getStoreId = function () { return this.storeId; };
    Cart.prototype.getStatus = function () { return this.status; };
    Cart.prototype.getCartItems = function () { return this.cartItems; };
    Cart.prototype.getCartCost = function () { return this.cartCost; };
    Cart.prototype.getTableId = function () { return this.tableId; };
    return Cart;
}());
exports.Cart = Cart;
//# sourceMappingURL=cart.js.map